
void push(uint8_t v){
	int sp;
	sp=ram[ports[PO_SPL]];
	if(ports[PO_SPH]) sp+=ram[ports[PO_SPH]]<<8;
	wrRam(sp,v);
	sp--;
	if(ports[PO_SPH]) ram[ports[PO_SPH]]=sp>>8;
	ram[ports[PO_SPL]]=sp&0xff;
}

uint8_t pop(void){
	uint8_t r;
	int sp;
	sp=ram[ports[PO_SPL]];
	if(ports[PO_SPH]) sp+=(ram[ports[PO_SPH]])<<8;
	sp++;
	r=rdRam(sp);
	if(ports[PO_SPH]) wrRam(ports[PO_SPH],sp>>8);
	wrRam(ports[PO_SPL],sp&0xff);
	return r;
}

/**
 * \brief überspringt einen Befehl
 */
void skip(void){
	int opc,mask,pat,nr,cy,c=0;
	opc=flash[pc].opcode;
	for(;;){
		mask=skip_tab[c++];
		pat=skip_tab[c++];
		nr=skip_tab[c++];
		if(!nr){
			/* nicht gefunden: gehen wir von 1 aus */
			pc++;
			ClockCycle();
			return;
		}
		if((opc&mask)==pat){
			pc+=nr;
			while(nr--) ClockCycle();
			return;
		}
	}
}

/**
 * \brief beschreibt ein Doppelregister mit einem Word
 */
void setPtr(int idx, int val){
	wrRam(idx,val&0xff);
	wrRam(idx+1,val>>8);
}

/**
 * \brief liefer den Wert eines Doppelregisters (X,Y oder X)
 * \param Index des geraden (ersten) Registers
 * \return Wert des Registers
 */
int getPtr(int idx){
	return rdRam(idx)+(rdRam(idx+1)<<8);
}

/**
 * \brief liefer den Wert eines Doppelregisters (X+,Y+ oder Z+),
 * anschließend wird es inkrementiert
 * \param Index des geraden (ersten) Registers
 * \return Wert des Registers vor der Inkrementierung
 */
int incPtr(int idx){
	int r;
	r=rdRam(idx)+(rdRam(idx+1)<<8);
	r++;
	wrRam(idx,r&0xff);
	wrRam(idx+1,r>>8);;
	return r-1;
}

/**
 * \brief liefer den Wert eines Doppelregisters (-X,-Y oder -Z),
 * welches vorher dekrementiert wurde
 * \param Index des geraden (ersten) Registers
 * \return Wert des Registers nach der Dekrementierung
 */
int decPtr(int idx){
	int r;
	r=rdRam(idx)+(rdRam(idx+1)<<8);
	r--;
	wrRam(idx,r&0xff);
	wrRam(idx+1,r>>8);;
	return r;
}

int getBit(uint16_t adr, int bit){
	uint8_t r;
	r=rdRam(adr);
	r>>=bit;
	return	r&1;
}

void setBit(uint16_t adr, int bit, int val){
	uint8_t r;
	r=rdRam(adr);
	if(val&1){
		(r)|=1<<bit;
	}
	else{
		(r)&=~(1<<bit);
	}
	wrRam(adr,r);
}

/**
 * \brief führt einen Prozessorbefehl aus
 */
void sim(void){
	int a,A,b,d,r,k,q,opc,V,N,C,S,H;
	opc=flash[pc].opcode;
	pc++;
	ClockCycle();

	if (opc==0x0000){ /* nop */
		return;
	}

	if (opc==0x9598){ /* break */
		return;
	}

	if (opc==0x9488){ /* clc */
		ram[ports[PO_SREG]]&=~(1<<fC);
		return;
	}

	if (opc==0x94D8){ /* clh */
		ram[ports[PO_SREG]]&=~(1<<fH);
		return;
	}

	if (opc==0x94F8){ /* cli */
		ram[ports[PO_SREG]]&=~(1<<fI);
		return;
	}

	if (opc==0x94A8){ /* cln */
		ram[ports[PO_SREG]]&=~(1<<fN);
		return;
	}

	if (opc==0x94C8){ /* cls */
		ram[ports[PO_SREG]]&=~(1<<fS);
		return;
	}

	if (opc==0x94E8){ /* clt */
		ram[ports[PO_SREG]]&=~(1<<fT);
		return;
	}

	if (opc==0x94B8){ /* clv */
		ram[ports[PO_SREG]]&=~(1<<fV);
		return;
	}

	if (opc==0x9498){ /* clz */
		ram[ports[PO_SREG]]&=~(1<<fZ);
		return;
	}

	if (opc==0x9519){ /* eicall */
		// sprintf(st,"eicall");
		return;
	}

	if (opc==0x9419){ /* eijmp */
		// sprintf(st,"eijmp");
		return;
	}

	if (opc==0x9509){ /* icall */
		push(pc&0xff);
		push(pc>>8);
		pc=getPtr(rrZ);
		level++;
		ClockCycle();
		ClockCycle();
		return;
	}

	if (opc==0x9409){ /* ijmp */
		pc=getPtr(rrZ);
		ClockCycle();
		return;
	}

	if (opc==0x9508){ /* ret */
		A=pop()<<8;
		A+=pop();
		pc=A;
		level--;
		ClockCycle();
		ClockCycle();
		ClockCycle();
		return;
	}

	if (opc==0x9518){ /* reti */
		A=pop()<<8;
		A+=pop();
		pc=A;
		level--;
		ram[ports[PO_SREG]]|=(1<<fI);
		ClockCycle();
		ClockCycle();
		ClockCycle();
		return;
	}

	if (opc==0x9408){ /* sec */
		ram[ports[PO_SREG]]|=1<<fC;
		return;
	}

	if (opc==0x9458){ /* seh */
		ram[ports[PO_SREG]]|=1<<fH;
		return;
	}

	if (opc==0x9478){ /* sei */
		ram[ports[PO_SREG]]|=1<<fI;
		return;
	}

	if (opc==0x9428){ /* sen */
		ram[ports[PO_SREG]]|=1<<fN;
		return;
	}

	if (opc==0x9448){ /* ses */
		ram[ports[PO_SREG]]|=1<<fS;
		return;
	}

	if (opc==0x9468){ /* set */
		ram[ports[PO_SREG]]|=1<<fT;
		return;
	}

	if (opc==0x9438){ /* sev */
		ram[ports[PO_SREG]]|=1<<fV;
		return;
	}

	if (opc==0x9418){ /* sez */
		ram[ports[PO_SREG]]|=1<<fZ;
		return;
	}

	if (opc==0x9588){ /* sleep */
		return;
	}

	if (opc==0x95E8){ /* spm */
		/* ToDo */
		return;
	}

	if (opc==0x95A8){ /* wdr */
		return;
	}

	if ((opc&0xFF0F)==0xEF0F){ /* ser */
		d=(opc>>4)&0x0f;
		ram[d+16]=0xff;
		return;
	}

	if ((opc&0xFE0F)==0x9400){ /* com */
		d=(opc>>4)&0x1f;
		A=rdRam(d);
		A=0xff-A;
		ram[d]=A;
		setBit(ports[PO_SREG],fZ,(A==0));
		N=(A>>7);
		setBit(ports[PO_SREG],fS,N);
		setBit(ports[PO_SREG],fN,N);
		ram[ports[PO_SREG]]&=~(1<<fV);
		ram[ports[PO_SREG]]|=(1<<fC);
		return;
	}

	if ((opc&0xFE0F)==0x940A){ /* dec */
		d=(opc>>4)&0x1f;
		A=rdRam(d);
		A--;
		A&=0xff;
		ram[d]=A;
		V=(A^0x80)==0xff;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fZ,(A==0));
		return;
	}

	if ((opc&0xFE0F)==0x9403){ /* inc */
		d=(opc>>4)&0x1f;
		A=rdRam(d);
		A++;
		A&=0xff;
		ram[d]=A;
		V=(A^0x80)==0xff;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fZ,(A==0));
		return;
	}

	if ((opc&0xFE0F)==0x9000){ /* lds */
		d=(opc>>4)&0x1f;
		k=flash[pc].opcode;
		pc++;
		ram[d]=rdRam(k);
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9200){ /* sts */
		d=(opc>>4)&0x1f;
		k=flash[pc].opcode;
		pc++;
		wrRam(k,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x900F){ /* pop */
		d=(opc>>4)&0x1f;
		ram[d]=pop();
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x920F){ /* push */
		d=(opc>>4)&0x1f;
		push(ram[d]);
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9401){ /* neg */
		d=(opc>>4)&0x1f;
		k=rdRam(d);
		A=0;
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		ram[d]=A;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xFE0F)==0x9406){ /* lsr */
		d=(opc>>4)&0x1f;
		A=rdRam(d);
		C=A&1;
		A>>=1;
		ram[d]=A;
		N=0;
		V=N^C;
		S=N^V;
		setBit(ports[PO_SREG],fC,C);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fS,S);
		return;
	}

	if ((opc&0xFE0F)==0x9405){ /* asr */
		d=(opc>>4)&0x1f;
		A=rdRam(d);
		C=(A&1);
		if(A&0x80) A|=0x100;
		A>>=1;
		ram[d]=A;
		N=(A>>7)&1;
		V=N^C;
		S=N^V;
		setBit(ports[PO_SREG],fC,C);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fS,S);
		return;
	}

	if ((opc&0xFE0F)==0x9407){ /* ror */
		d=(opc>>4)&0x1f;
		A=rdRam(d);
		if((ram[ports[PO_SREG]])&(1<<fC)) A|=0x100;
		C=A&1;
		A>>=1;
		ram[d]=A;
		N=A>>7;
		V=N^C;
		S=N^V;
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xFE0F)==0x9402){ /* swap */
		d=(opc>>4)&0x1f;
		k=rdRam(d);
		k<<=4;
		k|=k>>8;
		ram[d]=k&0xff;
		return;
	}

	if ((opc&0xFE0E)==0x940E){ /* call */
		k=((opc>>3)&0x3e)+(opc&1);
		k<<=16;
		k+=flash[pc].opcode;
		pc++;
		push(pc&0xff);
		push(pc>>8);
		pc=k;
		level++;
		ClockCycle();
		ClockCycle();
		ClockCycle();
		return;
	}

	if ((opc&0xFE0E)==0x940C){ /* jmp */
		k=((opc>>3)&0x3e)+(opc&1);
		k<<=16;
		k+=flash[pc].opcode;
		pc++;
		pc=k;
		ClockCycle();
		ClockCycle();
		return;
	}

	if ((opc&0xFF88)==0x0308){ /* fmul */
		d=((opc>>4)&0x7)+16;
		r=(opc&7)+16;
		A=(ram[d]*ram[r])<<1;
		ram[0]=A&0xff;
		ram[1]=A>>8;
		C=A>>15;
		setBit(ports[PO_SREG],1<<fZ,(A==0));
		setBit(ports[PO_SREG],1<<fC,C);
		return;
	}

	if ((opc&0xFF88)==0x0380){ /* fmuls */
		d=((opc>>4)&0x7)+16;
		r=(opc&7)+16;
		A=((uint8_t)ram[d]*(uint8_t)ram[r])<<1;
		ram[0]=A&0xff;
		ram[1]=A>>8;
		C=A>>15;
		setBit(ports[PO_SREG],1<<fZ,(A==0));
		setBit(ports[PO_SREG],1<<fC,C);
		return;
	}

	if ((opc&0xFF88)==0x0388){ /* fmulsu */
		d=((opc>>4)&0x7)+16;
		r=(opc&7)+16;
		A=((uint8_t)ram[d]*ram[r])<<1;
		ram[0]=A&0xff;
		ram[1]=A>>8;
		C=A>>15;
		setBit(ports[PO_SREG],1<<fZ,(A==0));
		setBit(ports[PO_SREG],1<<fC,C);
		return;
	}

	if ((opc&0xFF88)==0x0300){ /* mulsu */
		d=((opc>>4)&0x7)+16;
		r=opc&7;
		A=(int8_t)ram[d]*ram[r];
		ram[0]=A&0xff;
		ram[1]=A>>8;
		C=A>>15;
		setBit(ports[PO_SREG],1<<fZ,(A==0));
		setBit(ports[PO_SREG],1<<fC,C);
		return;
	}

	if ((opc&0xFF00)==0x0200){ /* muls */
		d=((opc>>4)&0x7)+16;
		r=opc&0x0f;
		A=(int8_t)ram[d]*(int8_t)ram[r];
		ram[0]=A&0xff;
		ram[1]=A>>8;
		C=A>>15;
		setBit(ports[PO_SREG],1<<fZ,(A==0));
		setBit(ports[PO_SREG],1<<fC,C);
		return;
	}

	if ((opc&0xFC00)==0x9C00){ /* mul */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		A=ram[d]*ram[r];
		ram[0]=A&0xff;
		ram[1]=A>>8;
		C=A>>15;
		setBit(ports[PO_SREG],1<<fZ,(A==0));
		setBit(ports[PO_SREG],1<<fC,C);
		return;
	}

	if ((opc&0xFC07)==0xF000){ /* brcs */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(((ram[ports[PO_SREG]])&(1<<fC))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF400){ /* brcc */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(!((ram[ports[PO_SREG]])&(1<<fC))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF001){ /* breq */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(((ram[ports[PO_SREG]])&(1<<fZ))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF401){ /* brne */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(!((ram[ports[PO_SREG]])&(1<<fZ))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF005){ /* brhs */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(((ram[ports[PO_SREG]])&(1<<fH))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF405){ /* brhc */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(!((ram[ports[PO_SREG]])&(1<<fH))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF006){ /* brts */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(((ram[ports[PO_SREG]])&(1<<fT))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF406){ /* brtc */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(!((ram[ports[PO_SREG]])&(1<<fT))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF003){ /* brvs */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(((ram[ports[PO_SREG]])&(1<<fV))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF403){ /* brvc */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(!((ram[ports[PO_SREG]])&(1<<fV))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF007){ /* brie */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(((ram[ports[PO_SREG]])&(1<<fI))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF407){ /* brid */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(!((ram[ports[PO_SREG]])&(1<<fI))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF002){ /* brmi */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(((ram[ports[PO_SREG]])&(1<<fN))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF402){ /* brpl */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(!((ram[ports[PO_SREG]])&(1<<fN))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF004){ /* brlt */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(((ram[ports[PO_SREG]])&(1<<fS))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFC07)==0xF404){ /* brge */
		k=(opc>>3)&0x7f;
		if(k>63) k-=128;
		if(!((ram[ports[PO_SREG]])&(1<<fS))){
			pc+=k;
			ClockCycle();
		}
		return;
	}

	if ((opc&0xFE08)==0xF800){ /* bld */
		d=(opc>>4)&0x1f;
		b=opc&7;
		A=1;
		if(b) A<<=b;
		if(ram[ports[PO_SREG]]&(1<<fT)) ram[d]|=A; else ram[d]&=~A;
		return;
	}

	if ((opc&0xFE08)==0xFA00){ /* bst */
		d=(opc>>4)&0x1f;
		b=opc&7;
		A=rdRam(d);
		if(b) A>>=b;
		if(A&1) ram[ports[PO_SREG]]|=(1<<fT); else ram[ports[PO_SREG]]&=~(1<<fT);
		return;
	}

	if ((opc&0xFF00)==0x9A00){ /* sbi */
		A=(opc>>3)&0x1f;
		b=opc&7;
		d=rdRam(0x20+A);
		d|=(1<<b);
		wrRam(0x20+A,d);
		return;
	}

	if ((opc&0xFF00)==0x9800){ /* cbi */
		A=(opc>>3)&0x1f;
		b=opc&7;
		d=rdRam(0x20+A);
		d&=~(1<<b);
		wrRam(0x20+A,d);
		return;
	}

	if ((opc&0xFF00)==0x0100){ /* movw */
		d=(opc>>4)&0x0f;
		r=opc&0x0f;
		A=getPtr(r<<1);
		setPtr(d<<1,A);
		return;
	}

	if ((opc&0xFF00)==0x9900){ /* sbic */
		A=(opc>>3)&0x1f;
		b=opc&7;
		if(!getBit(0x20+A,b)) skip();
		return;
	}

	if ((opc&0xFF00)==0x9B00){ /* sbis */
		A=(opc>>3)&0x1f;
		b=opc&7;
		if(ram[0x20+A]&d) skip();
		if(getBit(0x20+A,b)) skip();
		return;
	}

	if ((opc&0xFF00)==0x9700){ /* sbiw */
		k=((opc>>2)&0x30)+(opc&0x0f);
		d=((opc>>4)&3)*2+24;
		A=ram[d]+(ram[d+1]<<8);
		a=(A&0x7fff)-(k&0x7fff);/* fürs Overflow */
		V=a>>15;
		A=A-k;
		ram[d]=A&0xff;
		ram[d+1]=A>>8;
		C=A>>16;
		V^=C;
		N=(A>>15);
		S=N^V;
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);

		// sprintf(st,"sbiw r%d,%d",d+24,k);
		return;
	}

	if ((opc&0xFF00)==0x9600){ /* adiw */
		k=((opc>>2)&0x30)+(opc&0x0f);
		d=((opc>>4)&3)*2+24;
		A=ram[d]+(ram[d+1]<<8);
		a=(A&0x7fff)+(k&0x7fff);/* fürs Overflow */
		V=a>>15;
		A=A+k;
		ram[d]=A&0xff;
		ram[d+1]=A>>8;
		C=A>>16;
		V^=C;
		N=(A>>15);
		S=N^V;
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xFE08)==0xFC00){ /* sbrc */
		r=(opc>>4)&0x1f;
		b=opc&7;
		if(!getBit(r,b)) skip();
		return;
	}

	if ((opc&0xFE08)==0xFE00){ /* sbrs */
		r=(opc>>4)&0x1f;
		b=opc&7;
		if(getBit(r,b)) skip();
		return;
	}

	if (opc==0x95c8)		{ /* lpm r0,z */
		A=getPtr(rrZ);
		q=A&1;	/* oberes oder unteres? */
		A>>=1;
		b=flash[A].opcode;
		if(q) b>>=8; else b&=0xff;
		ram[0]=b;
		ClockCycle();
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9004){ /* lpm z */
		d=(opc>>4)&0x1f;
		A=getPtr(rrZ);
		q=A&1;	/* oberes oder unteres? */
		A>>=1;
		b=flash[A].opcode;
		if(q) b>>=8; else b&=0xff;
		ram[d]=b;
		ClockCycle();
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9005){ /* lpm z+ */
		d=(opc>>4)&0x1f;
		A=incPtr(rrZ);
		q=A&1;	/* oberes oder unteres? */
		A>>=1;
		b=flash[A].opcode;
		if(q) b>>=8; else b&=0xff;
		ram[d]=b;
		ClockCycle();
		ClockCycle();
		return;
	}

	if ((opc&0xFC00)==0x2400){ /* eor */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		A=rdRam(d);
		A^=rdRam(r);
		ram[d]=A;
		V=0;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fZ,(A==0));
		return;
	}

	if ((opc&0xFC00)==0x2800){ /* or */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		A=rdRam(d);
		A|=rdRam(r);
		ram[d]=A;
		V=0;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fZ,(A==0));
		return;
	}

	if ((opc&0xFC00)==0x2000){ /* and */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		A=rdRam(d);
		A&=rdRam(r);
		ram[d]=A;
		V=0;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fZ,(A==0));
		return;
	}

	if ((opc&0xFC00)==0x0800){ /* sbc */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		k=rdRam(r);
		if(ram[ports[PO_SREG]]&(1<<fC)) k++;
		A=rdRam(d);
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		ram[d]=A;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		if((ram[ports[PO_SREG]])&fZ) setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xFC00)==0x1800){ /* sub */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		k=rdRam(r);
		A=rdRam(d);
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		ram[d]=A;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xFC00)==0x0400){ /* cpc */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		k=rdRam(r);
		if(ram[ports[PO_SREG]]&(1<<fC)) k++;
		A=rdRam(d);
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		if((ram[ports[PO_SREG]])&fZ) setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xFC00)==0x1400){ /* cp */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		k=rdRam(r);
		A=rdRam(d);
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xFC00)==0x1000){ /* cpse */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		k=rdRam(r);
		A=rdRam(d);
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		if((ram[ports[PO_SREG]])&fZ) skip();
		return;
	}

	if ((opc&0xFC00)==0x1C00){ /* adc */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		k=rdRam(r);
		if(ram[ports[PO_SREG]]&(1<<fC)) k++;
		A=rdRam(d);
		a=(A&0xf)+(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)+(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A+k;
		ram[d]=A;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xFC00)==0x0C00){ /* add */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		k=rdRam(r);
		A=rdRam(d);
		a=(A&0xf)+(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)+(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A+k;
		ram[d]=A;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if (opc==0x95d8){ /* elpm1 r0,z*/
		d=(opc>>4)&0x1f;
		// sprintf(st,"elpm r0,z");
		return;
	}

	if ((opc&0xFE0F)==0x9006){ /* elpm z */
		d=(opc>>4)&0x1f;
		// sprintf(st,"elpm r%d,z",d);
		return;
	}

	if ((opc&0xFE0F)==0x9007){ /* elpm z+ */
		d=(opc>>4)&0x1f;
		// sprintf(st,"elpm r%d,z+",d);
		return;
	}

	if ((opc&0xFE0F)==0x900C){ /* ld x */
		d=(opc>>4)&0x1f;
		A=getPtr(rrX);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x900D){ /* ld x+ */
		d=(opc>>4)&0x1f;
		A=incPtr(rrX);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x900E){ /* ld -x */
		d=(opc>>4)&0x1f;
		A=decPtr(rrX);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x8008){ /* ld y */
		d=(opc>>4)&0x1f;
		A=getPtr(rrY);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9009){ /* ld y+ */
		d=(opc>>4)&0x1f;
		A=incPtr(rrY);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x900A){ /* ld -y */
		d=(opc>>4)&0x1f;
		A=decPtr(rrY);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xD208)==0x8008){ /* ldd y */
		d=(opc>>4)&0x1f;
		q=((opc>>8)&0x20)+((opc>>7)&0x18)+(opc&7);
		A=getPtr(rrY);
		wrRam(d,rdRam(A+q));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x8000){ /* ld z */
		d=(opc>>4)&0x1f;
		A=getPtr(rrZ);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9001){ /* ld z+ */
		d=(opc>>4)&0x1f;
		A=incPtr(rrZ);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9002){ /* ld -z */
		d=(opc>>4)&0x1f;
		A=decPtr(rrZ);
		wrRam(d,rdRam(A));
		ClockCycle();
		return;
	}

	if ((opc&0xD208)==0x8000){ /* ldd z */
		d=(opc>>4)&0x1f;
		q=((opc>>8)&0x20)+((opc>>7)&0x18)+(opc&7);
		A=getPtr(rrZ);
		wrRam(d,rdRam(A+q));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x920C){ /* st x */
		d=(opc>>4)&0x1f;
		A=getPtr(rrX);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x920D){ /* st x+ */
		d=(opc>>4)&0x1f;
		A=incPtr(rrX);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x920E){ /* st -x */
		d=(opc>>4)&0x1f;
		A=decPtr(rrX);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x8208){ /* st y */
		d=(opc>>4)&0x1f;
		A=getPtr(rrY);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9209){ /* st y+ */
		d=(opc>>4)&0x1f;
		A=incPtr(rrY);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x920A){ /* st -y */
		d=(opc>>4)&0x1f;
		A=decPtr(rrY);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xD208)==0x8208){ /* std y */
		d=(opc>>4)&0x1f;
		q=((opc>>8)&0x20)+((opc>>7)&0x18)+(opc&7);
		A=getPtr(rrY);
		wrRam(A+q,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x8200){ /* st z */
		d=(opc>>4)&0x1f;
		A=getPtr(rrZ);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9201){ /* st z+ */
		d=(opc>>4)&0x1f;
		A=incPtr(rrZ);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFE0F)==0x9202){ /* st -z */
		d=(opc>>4)&0x1f;
		A=decPtr(rrZ);
		wrRam(A,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xD208)==0x8200){ /* std z */
		d=(opc>>4)&0x1f;
		q=((opc>>8)&0x20)+((opc>>7)&0x18)+(opc&7);
		A=getPtr(rrZ);
		wrRam(A+q,rdRam(d));
		ClockCycle();
		return;
	}

	if ((opc&0xFC00)==0x2C00){ /* mov */
		r=((opc>>5)&0x10)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		wrRam(r,rdRam(d));
		return;
	}

	if ((opc&0xF800)==0xB000){ /* in */
		A=((opc>>5)&0x30)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		ram[d]=rdRam(A+32);
		return;
	}

	if ((opc&0xF800)==0xB800){ /* out */
		A=((opc>>5)&0x30)+(opc&0x0f);
		d=(opc>>4)&0x01f;
		wrRam(A+32,rdRam(d));
		return;
	}

	if ((opc&0xF000)==0x7000){ /* andi */
		k=((opc>>4)&0xf0)+(opc&0x0f);
		d=(opc>>4)&0x0f;
		d+=16;
		A=rdRam(d);
		A&=k;
		ram[d]=A;
		V=0;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fZ,(A==0));
		return;
	}

	if ((opc&0xF000)==0x3000){ /* cpi */
		k=((opc>>4)&0xf0)+(opc&0x0f);
		d=(opc>>4)&0x0f;
		d+=16;
		A=rdRam(d);
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xF000)==0xE000){ /* ldi */
		k=((opc>>4)&0xf0)+(opc&0x0f);
		d=(opc>>4)&0x0f;
		ram[d+16]=k;
		return;
	}

	if ((opc&0xF000)==0x6000){ /* ori */
		k=((opc>>4)&0xf0)+(opc&0x0f);
		d=(opc>>4)&0x0f;
		d+=16;
		A=rdRam(d);
		A|=k;
		ram[d]=A;
		V=0;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fZ,(A==0));
		return;
	}

	if ((opc&0xF000)==0xD000){ /* rcall */
		k=opc&0xfff;
		if(k>2047) k-=4096;
		push(pc&0xff);
		push(pc>>8);
		pc+=k;
		level++;
		ClockCycle();
		ClockCycle();
		return;
	}

	if ((opc&0xF000)==0xC000){ /* rjmp */
		k=opc&0xfff;
		if(k>2047) k-=4096;
		pc+=k;
		ClockCycle();
		return;
	}

	if ((opc&0xF000)==0x4000){ /* sbci */
		k=((opc>>4)&0xf0)+(opc&0x0f);
		d=(opc>>4)&0x0f;
		d+=16;
		if(ram[ports[PO_SREG]]&(1<<fC)) k++;
		A=rdRam(d);
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		ram[d]=A;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		if((ram[ports[PO_SREG]])&fZ) setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

	if ((opc&0xF000)==0x5000){ /* subi */
		k=((opc>>4)&0xf0)+(opc&0x0f);
		d=(opc>>4)&0x0f;
		d+=16;
		A=rdRam(d);
		a=(A&0xf)-(k&0xf);	/* fürs HalfCarry */
		H=a>>4;
		a=(A&0x7f)-(k&0x7f);/* fürs Overflow */
		V=a>>7;
		A=A-k;
		ram[d]=A;
		C=A>>8;
		V^=C;
		N=(A>>7);
		S=N^V;
		setBit(ports[PO_SREG],fH,H);
		setBit(ports[PO_SREG],fS,S);
		setBit(ports[PO_SREG],fV,V);
		setBit(ports[PO_SREG],fN,N);
		setBit(ports[PO_SREG],fZ,(A==0));
		setBit(ports[PO_SREG],fC,C);
		return;
	}

		// sprintf(st,"illegal opcode %04X",opc);
		return;
}
